==> providence-testing-1.10.0-sources.jar: net/morimekta/providence/testing/EqualToMessage.java <==
/*
 * Copyright 2016 Providence Authors
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package net.morimekta.providence.testing;

import com.google.common.collect.ImmutableSet;
import junit.framework.AssertionFailedError;
import net.morimekta.providence.PEnumValue;
import net.morimekta.providence.PMessage;
import net.morimekta.providence.PMessageVariant;
import net.morimekta.providence.PUnion;
import net.morimekta.providence.descriptor.PField;
import net.morimekta.util.Binary;
import net.morimekta.util.Strings;
import org.hamcrest.BaseMatcher;
import org.hamcrest.Description;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * @author Stein Eldar Johnsen
 * @since 21.01.16.
 */
public class EqualToMessage<Message extends PMessage<Message, Field>, Field extends PField>
        extends BaseMatcher<Message> {
    private final Message expected;
    private final Set<PField> ignoringFields;

    public EqualToMessage(Message expected) {
        this.expected = expected;
        this.ignoringFields = ImmutableSet.of();
    }

    private EqualToMessage(Message expected, Set<PField> ignoringFields) {
        this.expected = expected;
        this.ignoringFields = ignoringFields;
    }

    public EqualToMessage<Message, Field> ignoring(PField... fields) {
        return new EqualToMessage<>(expected, ImmutableSet.copyOf(fields));
    }

    @Override
    @SuppressWarnings("unchecked")
    public boolean matches(Object actual) {
        if (expected == null) {
            return actual == null;
        }
        if (!(actual instanceof PMessage)) {
            throw new AssertionFailedError("Item " + actual.getClass().toString() + " not a providence message.");
        }
        ArrayList<String> mismatches = new ArrayList<>();
        collectMismatches("", (PMessage) expected, (PMessage) actual, mismatches);
        return mismatches.isEmpty();
    }

    @Override
    public void describeTo(Description description) {
        description.appendText("equals(")
                   .appendText(limitToString(expected))
                   .appendText(")");
    }

    @Override
    @SuppressWarnings("unchecked")
    public void describeMismatch(Object actual, Description mismatchDescription) {
        if (expected == null) {
            mismatchDescription.appendText("got " + toString(actual));
        } else if (actual == null) {
            mismatchDescription.appendText("got null");
        } else {
            ArrayList<String> mismatches = new ArrayList<>();
            collectMismatches("", (PMessage) expected, (PMessage) actual, mismatches);
            if (mismatches.size() == 1) {
                mismatchDescription.appendText(mismatches.get(0));
            } else {
                boolean first = true;
                mismatchDescription.appendText("[");
                int i = 0;
                for (String mismatch : mismatches) {
                    if (first) {
                        first = false;
                    } else {
                        mismatchDescription.appendText(",");
                    }
                    mismatchDescription.appendText("\n        ");
                    if (i >= 20) {
                        int remaining = mismatches.size() - i;
                        mismatchDescription.appendText("... and " + remaining + " more");
                        break;
                    }
                    mismatchDescription.appendText(mismatch);
                    ++i;
                }
                mismatchDescription.appendText("\n     ]");
            }
        }
    }

    private <T extends PMessage<T, F>, F extends PField>
    void collectMismatches(String xPath, T expected, T actual, ArrayList<String> mismatches) {
        // This is pretty heavy calculation, but since it's only done on
        // mismatch / test failure, it should be fine.
        if (expected.descriptor()
                    .getVariant() == PMessageVariant.UNION) {
            PUnion<?,?> eu = (PUnion) expected;
            PUnion<?,?> ac = (PUnion) actual;

            if (!eu.unionField()
                   .equals(ac.unionField())) {
                mismatches.add(String.format(Locale.US, "%s to have %s, but had %s",
                                             xPath,
                                             eu.unionField()
                                               .getName(),
                                             ac.unionField()
                                               .getName()));
            }
        }

        for (PField field : expected.descriptor().getFields()) {
            if (ignoringFields.contains(field)) {
                continue;
            }

            int key = field.getId();
            String fieldXPath = xPath.isEmpty() ? field.getName() : xPath + "." + field.getName();

            if (expected.has(key) != actual.has(key)) {
                if (!expected.has(key)) {
                    mismatches.add(String.format(Locale.US, "%s to be missing, but was %s",
                                                 fieldXPath,
                                                 toString(actual.get(field.getId()))));
                } else if (!actual.has(key)) {
                    mismatches.add(String.format(Locale.US, "%s to be %s, but was missing",
                                                 fieldXPath,
                                                 toString(expected.get(field.getId()))));
                }
            } else if (!Objects.equals(expected.get(key), actual.get(key))) {
                switch (field.getType()) {
                    case MESSAGE: {
                        collectMismatches(fieldXPath,
                                          expected.get(key),
                                          actual.get(key),
                                          mismatches);
                        break;
                    }
                    case LIST: {
                        collectListMismatches(fieldXPath, expected.get(key), actual.get(key), mismatches);
                        break;
                    }
                    case SET: {
                        collectSetMismatches(fieldXPath, expected.get(key), actual.get(key), mismatches);
                        break;
                    }
                    case MAP: {
                        collectMapMismatches(fieldXPath, expected.get(key), actual.get(key), mismatches);
                        break;
                    }
                    default: {
                        mismatches.add(String.format(Locale.US, "%s was %s, expected %s",
                                                     fieldXPath,
                                                     toString(actual.get(field.getId())),
                                                     toString(expected.get(field.getId()))));
                        break;
                    }
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private <K, V> void collectMapMismatches(String xPath,
                                             Map<K, V> expected,
                                             Map<K, V> actual,
                                             ArrayList<String> mismatches) {
        mismatches.addAll(actual.keySet()
                                .stream()
                                .filter(key -> !expected.keySet()
                                                        .contains(key))
                                .map(key -> String.format(Locale.US, "found unexpected entry (%s, %s) in %s",
                                                          toString(key),
                                                          toString(actual.get(key)),
                                                          xPath))
                                .collect(Collectors.toList()));

        for (Map.Entry<K,V> entry : expected.entrySet()) {
            if (!actual.keySet()
                       .contains(entry.getKey())) {
                mismatches.add(String.format(Locale.US, "did not find entry (%s, %s) in in %s",
                                             toString(entry.getKey()),
                                             toString(expected.get(entry.getKey())),
                                             xPath));
            } else {
                V exp = entry.getValue();
                V act = actual.get(entry.getKey());
                if (!Objects.equals(exp, act)) {
                    // value differs.
                    String keyedXPath = String.format(Locale.US, "%s[%s]", xPath, toString(entry));
                    if (exp == null || act == null) {
                        mismatches.add(String.format(Locale.US, "%s was %s, should be %s",
                                                     keyedXPath,
                                                     toString(exp),
                                                     toString(act)));
                    } else if (act instanceof PMessage) {
                        collectMismatches(keyedXPath, (PMessage) exp, (PMessage) act, mismatches);
                    } else {
                        mismatches.add(String.format(Locale.US, "%s was %s, should be %s",
                                                     keyedXPath,
                                                     toString(act),
                                                     toString(exp)));
                    }
                }
            }
        }
    }

    private static <T> void collectSetMismatches(String xPath,
                                                 Set<T> expected,
                                                 Set<T> actual,
                                                 ArrayList<String> mismatches) {
        // order does NOT matter regardless of type. The only
        // errors are missing and unexpected values. Partial
        // matches are not checked.
        mismatches.addAll(actual.stream()
                                .filter(item -> !expected.contains(item))
                                .map(item -> String.format(Locale.US, "found unexpected set value %s in %s",
                                                           toString(item),
                                                           xPath))
                                .collect(Collectors.toList()));

        mismatches.addAll(expected.stream()
                                  .filter(item -> !actual.contains(item))
                                  .map(item -> String.format(Locale.US, "did not find value %s in %s", toString(item), xPath))
                                  .collect(Collectors.toList()));

    }

    @SuppressWarnings("unchecked")
    private <T> void collectListMismatches(String xPath,
                                           List<T> expected,
                                           List<T> actual,
                                           ArrayList<String> mismatches) {
        Set<T> handledItems = new HashSet<>();

        boolean hasReorder = false;
        ArrayList<String> reordering = new ArrayList<>();
        for (int expectedIndex = 0; expectedIndex < expected.size(); ++expectedIndex) {
            String indexedXPath = String.format(Locale.US, "%s[%d]", xPath, expectedIndex);
            T expectedItem = expected.get(expectedIndex);
            handledItems.add(expectedItem);

            T actualItem = actual.size() > expectedIndex ? actual.get(expectedIndex) : null;
            if (Objects.equals(expectedItem, actualItem)) {
                continue;
            }
            int actualIndex = actual.indexOf(expectedItem);

            int actualItemExpectedIndex = -1;
            if (actualItem != null) {
                actualItemExpectedIndex = expected.indexOf(actualItem);
            }

            if (actualIndex < 0) {
                reordering.add("NaN");
                // this item is missing.
                if (actualItemExpectedIndex < 0) {
                    handledItems.add(actualItem);
                    // replaced with new item, diff them normally.
                    if (actualItem instanceof PMessage) {
                        collectMismatches(indexedXPath, (PMessage) expectedItem, (PMessage) actualItem, mismatches);
                    } else {
                        mismatches.add(String.format(Locale.US, "expected %s to be %s, but was %s",
                                                     indexedXPath,
                                                     toString(expectedItem),
                                                     toString(actualItem)));
                    }
                } else {
                    // the other item is reordered, so this is blindly inserted.
                    mismatches.add(String.format(Locale.US, "missing item %s in %s", toString(expectedItem), indexedXPath));
                }
            } else if (actualIndex != expectedIndex) {
                reordering.add(String.format(Locale.US, "%+d", actualIndex - expectedIndex));
                hasReorder = true;
            } else {
                reordering.add("±0");
            }
        }
        for (int actualIndex = 0; actualIndex < actual.size(); ++actualIndex) {
            T actualItem = actual.get(actualIndex);
            if (handledItems.contains(actualItem)) {
                continue;
            }
            if (expected.contains(actualItem)) {
                continue;
            }
            String indexedXPath = String.format(Locale.US, "%s[%d]", xPath, actualIndex);
            mismatches.add(String.format(Locale.US, "unexpected item %s in %s", toString(actualItem), indexedXPath));
        }
        if (hasReorder) {
            mismatches.add(String.format(Locale.US, "unexpected item ordering in %s: [%s]", xPath, Strings.join(",", reordering)));
        }

    }

    protected static String toString(Object o) {
        if (o == null) {
            return "null";
        } else if (o instanceof PMessage) {
            return limitToString((PMessage) o);
        } else if (o instanceof PEnumValue) {
            return ((PEnumValue) o).descriptor()
                                   .getName() + "." + ((PEnumValue) o).asString();
        } else if (o instanceof Map) {
            return "{" + Strings.join(",",
                                      ((Map<?, ?>) o).entrySet()
                                                     .stream()
                                                     .map(e -> toString(e.getKey()) + ":" + toString(e.getValue()))
                                                     .collect(Collectors.toList())) + "}";
        } else if (o instanceof Collection) {
            return "[" + Strings.join(",",
                                      ((Collection<?>) o).stream()
                                                         .map(EqualToMessage::toString)
                                                         .collect(Collectors.toList())) + "]";
        } else if (o instanceof CharSequence) {
            return "\"" + Strings.escape(o.toString()) + "\"";
        } else if (o instanceof Binary) {
            int len = ((Binary) o).length();
            if (len > 110) {
                return String.format(Locale.US, "binary[%s...+%d]",
                                     ((Binary) o).toHexString()
                                                 .substring(0, 100),
                                     len - 50);
            } else {
                return "binary[" + ((Binary) o).toHexString() + "]";
            }
        } else if (o instanceof Double) {
            long l = ((Double) o).longValue();
            if (o.equals((double) l)) {
                return Long.toString(l);
            } else {
                return o.toString();
            }
        } else {
            return o.toString();
        }
    }

    private static String limitToString(PMessage<?, ?> message) {
        String tos = message == null ? "null" : message.asString();
        if (tos.length() > 120) {
            tos = tos.substring(0, 110) + "...}";
        }

        return tos;
    }
}

==> providence-testing-2.2.1-sources.jar: net/morimekta/providence/testing/EqualToMessage.java <==
/*
 * Copyright 2016 Providence Authors
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package net.morimekta.providence.testing;

import net.morimekta.providence.PEnumValue;
import net.morimekta.providence.PMessage;
import net.morimekta.providence.PMessageOrBuilder;
import net.morimekta.providence.descriptor.PField;
import net.morimekta.util.Binary;
import net.morimekta.util.Strings;
import net.morimekta.util.collect.UnmodifiableSet;
import org.hamcrest.BaseMatcher;
import org.hamcrest.Description;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static net.morimekta.providence.testing.util.MessageDiff.collectMismatches;
import static net.morimekta.providence.testing.util.MessageDiff.limitToString;

/**
 * Equality matcher for messages with pinpointed field diff output and
 * possibility to ignore individual fields.
 */
public class EqualToMessage<Message extends PMessage<Message>>
        extends BaseMatcher<PMessageOrBuilder<Message>> {
    private final PMessageOrBuilder<Message> expected;
    private final Set<PField> ignoringFields;

    public static <Message extends PMessage<Message>>
    EqualToMessage<Message> equalToMessage(PMessageOrBuilder<Message> expected) {
        return new EqualToMessage<>(expected);
    }

    public EqualToMessage(PMessageOrBuilder<Message> expected) {
        this.expected = expected.toMessage();
        this.ignoringFields = UnmodifiableSet.setOf();
    }

    private EqualToMessage(PMessageOrBuilder<Message> expected, Set<PField> ignoringFields) {
        this.expected = expected.toMessage();
        this.ignoringFields = ignoringFields;
    }

    public final EqualToMessage<Message> ignoring(PField... fields) {
        return new EqualToMessage<>(expected, UnmodifiableSet.copyOf(fields));
    }

    @Override
    @SuppressWarnings("unchecked")
    public boolean matches(Object actual) {
        if (expected == null) {
            return actual == null;
        }
        if (!(actual instanceof PMessage)) {
            throw new AssertionError("Item " + actual.getClass().toString() + " not a providence message.");
        }
        if (!((PMessage) actual).descriptor().equals(expected.descriptor())) {
            throw new AssertionError("Expected message type " + expected.descriptor().getQualifiedName() +
                                     ", but got " + ((PMessage) actual).descriptor().getQualifiedName());
        }
        try {
            collectMismatches("", expected, (Message) actual, UnmodifiableSet.copyOf(ignoringFields), str -> {
                throw new IllegalStateException();
            });
            return true;
        } catch (IllegalStateException e) {
            return false;
        }
    }

    @Override
    public void describeTo(Description description) {
        description.appendText("equals(")
                   .appendText(limitToString(expected))
                   .appendText(")");
    }

    @Override
    @SuppressWarnings("unchecked")
    public void describeMismatch(Object actual, Description mismatchDescription) {
        if (expected == null) {
            mismatchDescription.appendText("got " + toString(actual));
        } else if (actual == null) {
            mismatchDescription.appendText("got null");
        } else {
            ArrayList<String> mismatches = new ArrayList<>();
            collectMismatches("", expected, (PMessage) actual, UnmodifiableSet.copyOf(ignoringFields), mismatches::add);
            if (mismatches.size() == 1) {
                mismatchDescription.appendText(mismatches.get(0));
            } else {
                boolean first = true;
                mismatchDescription.appendText("[");
                int i = 0;
                for (String mismatch : mismatches) {
                    if (first) {
                        first = false;
                    } else {
                        mismatchDescription.appendText(",");
                    }
                    mismatchDescription.appendText("\n        ");
                    if (i >= 20) {
                        int remaining = mismatches.size() - i;
                        mismatchDescription.appendText("... and " + remaining + " more");
                        break;
                    }
                    mismatchDescription.appendText(mismatch);
                    ++i;
                }
                mismatchDescription.appendText("\n     ]");
            }
        }
    }

    protected static String toString(Object o) {
        if (o == null) {
            return "null";
        } else if (o instanceof PMessage) {
            return limitToString((PMessage<?>) o);
        } else if (o instanceof PEnumValue) {
            return ((PEnumValue) o).descriptor()
                                   .getName() + "." + ((PEnumValue) o).asString();
        } else if (o instanceof Map) {
            return "{" + Strings.join(",",
                                      ((Map<?, ?>) o).entrySet()
                                                     .stream()
                                                     .map(e -> toString(e.getKey()) + ":" + toString(e.getValue()))
                                                     .collect(Collectors.toList())) + "}";
        } else if (o instanceof Collection) {
            return "[" + Strings.join(",",
                                      ((Collection<?>) o).stream()
                                                         .map(EqualToMessage::toString)
                                                         .collect(Collectors.toList())) + "]";
        } else if (o instanceof CharSequence) {
            return "\"" + Strings.escape(o.toString()) + "\"";
        } else if (o instanceof Binary) {
            int len = ((Binary) o).length();
            if (len > 110) {
                return String.format(Locale.US, "binary[%s...+%d]",
                                     ((Binary) o).toHexString()
                                                 .substring(0, 100),
                                     len - 50);
            } else {
                return "binary[" + ((Binary) o).toHexString() + "]";
            }
        } else if (o instanceof Double) {
            long l = ((Double) o).longValue();
            if (o.equals((double) l)) {
                return Long.toString(l);
            } else {
                return o.toString();
            }
        } else {
            return o.toString();
        }
    }
}
